:
# build.emacs -- make a stand-alone emacs binary

# set -x

do_strip=yes
do_opt=yes
do_debug=no
do_clean=no
do_profile=no
do_warn=no
do_pipe=
do_static=
is_gcc=no
is_bgcc=no
set_bgcc=no
curses=

if [ -z "$MAKE" ]
then
	if [ -x "/usr/local/bin/gmake" ]
	then MAKE="/usr/local/bin/gmake"
	else MAKE=make
	fi
fi

set_bgcc_flags(){
	is_gcc=yes ; is_bgcc=yes ; do_strip=no ; do_debug=yes ; do_opt=no
}

if [ ! -z "$DEF_CC" ]
then
	CC="$DEF_CC"
	export CC
fi

case "$DEF_CC" in
bgcc*)	set_bgcc_flags
	CC=bgcc
	export CC
	;;
*gcc*)	is_gcc=yes
	;;
esac

while [ "X$1" != "X" ]
do
	case "$1" in
	-s|-strip)	do_strip=yes ;;
	-sym)		do_strip=no ;;
	-g|-debug)	do_debug=yes ;;
	-O)		do_opt=yes ;;
	-p)		do_profile=yes ;;
	-w|-W3|-warn)	do_warn=yes ;;
	-pipe)		do_pipe=yes ;;
	-static)	do_static=yes ;;
	-clean)		do_clean=yes ;;
	-curses)	curses=curses ;;
	-ncurses)	curses=ncurses ;;
	-cursest)	curses="curses -ltermcap" ;;
	-termcap)	curses="termcap" ;;
	-bgcc)		if [ "$is_bgcc" != yes ]
			then
				set_bgcc=yes
				set_bgcc_flags
			fi
			;;
	-x)		set -x ;;
	-*)		echo "$0: Invalid option $1" ; exit ;;
	*)		echo "$0: Invalid parameter $1" ; exit ;;
	esac
	shift
done

if [ "$set_bgcc" = yes ]
then
	save_do_opt="$do_opt"
	save_do_strip="$do_strip"
	save_is_bgcc="$is_bgcc"
	save_is_gcc="$is_gcc"

	. setgcc.sh -bgcc

	do_opt="$save_do_opt"
	do_strip="$save_do_strip"
	is_bgcc="$save_is_bgcc"
	is_gcc="$save_is_gcc"

	CC="bgcc"
	export CC
fi

if [ "$do_debug" = yes ] ; then do_strip=no ; do_opt=no ; fi
if [ "$do_profile" = yes ] ; then do_strip=no ; fi

MEEXTRAFLAGS="$DEF_WARN"

if [ "$do_warn" = yes ]
then
	MEEXTRAFLAGS="$MEEXTRAFLAGS $FULL_WARN"
fi

if [ "$do_opt" = yes ]
then
	MEEXTRAFLAGS="$MEEXTRAFLAGS $FULL_OPT"
fi

if [ "$do_static" = yes ]
then
	MEEXTRAFLAGS="$MEEXTRAFLAGS -static"
fi

if [ ! -z "$SPICE_INC" -a ! -d "$SPICE_INC" ]
then
	echo "$0: SPICE_INC $SPICE_INC directory not found"
	SPICE_INC=
fi

if [ -z "$SPICE_INC" ]
then
	# Building without SCS spice environment
	MEEXTRAFLAGS="$MEEXTRAFLAGS -D__PORT_H__"
else
	# Building under SCS spice environment
	MEEXTRAFLAGS="$MEEXTRAFLAGS -I$SPICE_INC"
fi

case `uname -r` in
  3.2)	# options for SCO Unix 3.2v4
	if [ "$is_gcc" = no ]
	then
		case `uname -s` in
		SCO_SV)
			;;
		*)
			MEEXTRAFLAGS="$MEEXTRAFLAGS -nointl"
			if [ $do_opt = yes ]
			then MEEXTRAFLAGS="$MEEXTRAFLAGS -Otacl -CSON" ; fi
			;;
		esac
	fi
	;;
esac

case `uname -s` in
UNIX_SV) # options for SCO UnixWare 2.1.2
	MEEXTRAFLAGS="$MEEXTRAFLAGS -D_SYSTYPE_SVR4=1"
	if [ "$do_opt" = yes ]
	then MEEXTRAFLAGS="$MEEXTRAFLAGS -O -Khost,fixed_frame,inline,pentium" ; fi
	;;
FreeBSD|Linux)
	do_pipe=yes
	if [ "$do_opt" = yes ]
	then MEEXTRAFLAGS="$MEEXTRAFLAGS -O3 -fomit-frame-pointer -ffast-math" ; fi
	;;
Darwin)
	do_strip=no
	do_pipe=yes
	if [ "$do_opt" = yes ]
	then MEEXTRAFLAGS="$MEEXTRAFLAGS -O3 -fomit-frame-pointer -ffast-math" ; fi
	;;
HP-UX)
	case `uname -m` in
	ia64) if [ -z "$LIB_PREFIX" -a -f /usr/lib/hpux64/libc.so ] ; then LIB_PREFIX="/usr/lib/hpux64/lib" ; fi
	esac
	;;
esac

if [ -z "$LIB_PREFIX" ]
then
	if [ -f "/lib/libc.a" -o -f "/lib/libc_s" ]
	then
		LIB_PREFIX="/lib/lib"
	elif [ -d "/cygnus/cygwin-b20/H-i586-cygwin32/lib" ]
	then
		LIB_PREFIX="/cygnus/cygwin-b20/H-i586-cygwin32/lib/lib"
	elif [ -d /lib/386 ]
	then
		LIB_PREFIX="/lib/386/Slib"
	else
		LIB_PREFIX="/lib/lib"
	fi
fi

# MEEXTRALIB="-ltermcap -lmalloc -lc_s -lc -lx -lc_s"

MEEXTRALIB=

if [ -z "$curses" ]
then
	if [ -f ${LIB_PREFIX}ncurses.a -o -f /usr${LIB_PREFIX}ncurses.a -o \
		-f ${LIB_PREFIX}ncurses.dylib -o -f /usr${LIB_PREFIX}ncurses.dylib ]
	then
		# Use the ncurses library
		curses=ncurses
	fi
fi

if [ -z "$curses" ]
then
	if [ -f ${LIB_PREFIX}termcap.a -o -f /usr${LIB_PREFIX}termcap.a ]
	then
		# Use the termcap library
		curses=termcap
	else
		case "$LD_LIBRARY_PATH" in
		*:/usr/ucblib:*)
			if [ -f /usr/ucblib/libtermcap.a ]
			then
				# Use the termcap compatibility library
				curses=termcap
			fi
			;;
		esac
	fi
fi

if [ -z "$curses" ]
then
	if [ -f ${LIB_PREFIX}curses.a -o -f ${LIB_PREFIX}curses.so -o -f /usr/ccs${LIB_PREFIX}curses.a ]
	then
		# Use the curses library
		curses=curses
	fi
fi

if [ ! -z "$curses" ]
then
	MEEXTRALIB="$MEEXTRALIB -l$curses"
fi

if [ "$is_bgcc" = yes ]
then
	# Use bounds checking helper library (not needed for Linux)
	# MEEXTRALIB="$MEEXTRALIB -lcheck"
	x=x
else
	if [ -f ${LIB_PREFIX}malloc.a -o -f /usr${LIB_PREFIX}malloc.a ]
	then
		# Use fast malloc library
		MEEXTRALIB="$MEEXTRALIB -lmalloc"
	fi

	if [ $do_debug = no -a $do_strip = yes -a -f ${LIB_PREFIX}c_s.a ]
	then
		# Use shared version of c runline library
		MEEXTRALIB="$MEEXTRALIB -lc_s"
	fi
fi

if [ -f ${LIB_PREFIX}x.a ]
then
	# SCO needs -lx for nap call
	MEEXTRALIB="$MEEXTRALIB -lc -lx"
fi

#if [ -f /usr/ucblib/libucb.a ]
#then
#	if [ "`uname`" = SunOS ]
#	then
#		# Solaris needs ucb for usleep
#		# Do this last so we do not get more ucb than we want
#		MEEXTRALIB="$MEEXTRALIB -lc /usr/ucblib/libucb.a"
#	fi
#fi

MEEXTRALIB="$MEEXTRALIB $DEF_LIB"

if [ "$is_bgcc" = yes -a ! -z "$MEEXTRALIB" ]
then
	new_lib=
	for name in $MEEXTRALIB
	do
		case "$name" in
		-lmalloc|-lc|-lc_s)	;;
		*)		new_lib="$new_lib $name"
		esac
	done
	MEEXTRALIB="$new_lib"
fi		

if [ "$do_opt" = yes ] ; then MEEXTRAFLAGS="-O $MEEXTRAFLAGS" ; fi
if [ "$do_strip" = yes ] ; then MEEXTRAFLAGS="$MEEXTRAFLAGS -s" ; fi
if [ "$do_debug" = yes ] ; then MEEXTRAFLAGS="$MEEXTRAFLAGS -g" ; fi
if [ "$do_profile" = yes ] ; then MEEXTRAFLAGS="$MEEXTRAFLAGS -p" ; fi
if [ "$do_pipe" = yes ] ; then MEEXTRAFLAGS="$MEEXTRAFLAGS -pipe" ; fi

export MEEXTRAFLAGS MEEXTRALIB

echo "Building emacs with $MAKE"
echo "Compile flags: $MEEXTRAFLAGS"
echo "Link flags: $MEEXTRALIB"

if [ $do_clean = yes ]
then
	$MAKE -f mymakefile clean
fi

echo
echo "Warnings OK for"
echo "display.c: 1905: mlwrite has variable arguments"
echo "fileio.c: 383: long/short mismatch in argument to open"
echo "termio.c: 242: long/short mismatch in argument to kill"
echo

$MAKE -f mymakefile

ls -l emacs

if [ $do_strip = yes -a -x emacs -a "$is_bgcc" = no ]
then
	if [ -x /usr/bin/strip -o -x /usr/ccs/bin/strip ]
	then
		echo "Stripping..."
		strip emacs
		ls -l emacs
	fi
	if [ -x /usr/bin/mcs -o -x /usr/ccs/bin/mcs ]
	then
		echo "Removing comments..."
		mcs -d emacs
		ls -l emacs
	fi
fi
